from pathlib import Path
from typing import Dict, List, Optional, Union
import yaml

import pandas as pd

from opa_tools.config import get_request_dir
from .transformation import flatten
from .utils import ALLOWED_KINDS, normalize_request_identifier, format_identifier


REQUEST_DIR = get_request_dir()


def _yaml_read(path: Path):
    with open(path, "r", encoding="utf-8") as rf:
        return yaml.safe_load(rf)

def _find_zero_padded_subdir(path: Path, subdir: str, max_digits=12):
    if len(subdir) >= max_digits:
        raise RuntimeError(
            f"Subdirectory {subdir} too long to be found in {path}"
        )
    for n_zeros in range(0, max_digits - len(subdir)):
        cur_path = path / subdir.zfill(n_zeros)
        if cur_path.exists():
            return cur_path
    raise FileNotFoundError(
        f"Directory {path / subdir} (or any other zero-padded variant, "
        f"up to {max_digits} digits) not found for request."
    )

def _get_subrequests(
        request_path: Path,
        no_exist_ok: bool = False
    ) -> Dict:
    raise NotImplementedError(
        "Ability to fetch lists of subrequests not implemented yet. "
        "Please, pass subrequests manually (syntax: YYMM/REQID.SUBREQID)"
    )


def _get_single_request(
        name: str,
        subrequest: Union[str, None],
        month: str,
        kind: str,
        no_exist_ok: bool = False
    ) -> Dict:
    if kind not in ALLOWED_KINDS:
        raise ValueError(
            f"kind must be either 'opa' or 'gsv', but was '{kind}'."
        )
    request_dir = _find_zero_padded_subdir(REQUEST_DIR / month, name)
    if subrequest:
        request_dir = _find_zero_padded_subdir(request_dir, subrequest)
    request_path = request_dir / f"{kind}.yml"
    if request_path.exists():
        return _yaml_read(request_path)
    if any(subpath.is_dir() for subpath in request_dir.iterdir()):
        return _get_subrequests(request_path, no_exist_ok=no_exist_ok)
    if no_exist_ok:
        return {}
    raise FileNotFoundError(
        f"Directory {request_dir} exists but does not contain the specified request(s)."
    )

def _subdir_iterator(path: Path, level=0, only_terminal=False):
    if level not in {0, 1}:
        raise NotImplementedError(
            "Level must be either 0 or 1."
        )
    for subdir in sorted(path.iterdir(), key=lambda p: p.name):
        if not subdir.is_dir():
            continue
        if not only_terminal:
            yield subdir
            continue
        is_terminal_subdir = not any(
            subsubdir.is_dir() for subsubdir in subdir.iterdir()
        )
        if is_terminal_subdir:
            yield subdir
            continue
        elif level > 0:
            yield from _subdir_iterator(subdir, level - 1, only_terminal=True)
            continue

def get_request_identifiers(month: str = "*", subrequests: bool = False):
    if month == "*":
        output = []
        for month in sorted(path.name for path in REQUEST_DIR.iterdir()):
            output.extend(get_request_identifiers(month=month, subrequests=subrequests))
        return output
    month_dir = REQUEST_DIR / month
    if not month_dir.exists():
        raise FileNotFoundError(
            f"Month directory {month_dir} does not exists."
        )
    output = []
    for subdir in _subdir_iterator(path=month_dir, level=1, only_terminal=subrequests):
        subdir_parts = subdir.relative_to(REQUEST_DIR).parts
        # TODO: Check that is really a month
        output.append(format_identifier(*subdir_parts))
    return output


def _get_requests(
    name: str,
    subrequest: Union[str, None],
    month: str,
    kind: str,
    flat: bool,

) -> List[Dict] | Dict:
    if kind == "*":
        no_exist_error = ("gsv", "opa")
        requests = {
            kind: _get_single_request(
                name=name,
                subrequest=subrequest,
                month=month,
                kind=kind,
                no_exist_ok=(kind not in no_exist_error)
            ) for kind in ALLOWED_KINDS
        }
        if flat:
            return flatten(**requests)
        return [requests[kind] for kind in ALLOWED_KINDS]
    return [
        _get_single_request(name=name, subrequest=subrequest, month=month, kind=kind)
    ]

def get_request(
    identifier: Optional[Union[str, int, float]] = None,
    /,
    subrequest: Optional[str] = None,
    month: Optional[str] = None,
    kind: str = "*",
    *,
    flat=False,
    single_as_list=False,
) -> Union[List[Dict], Dict]:
    if identifier is None:
        if month is None:
            month = "*"
        output = []
        for identifier in get_request_identifiers(month=month):
            output.append(
                get_request(identifier, kind=kind, flat=flat, single_as_list=True)
            )
        return output
    name, subrequest, month, kind = normalize_request_identifier(
        identifier, subrequest=subrequest, month=month, kind=kind
    )
    requests = _get_requests(
        name=name, subrequest=subrequest, month=month, kind=kind, flat=flat
    )
    if len(requests) == 1 and not single_as_list:
        return requests[0]
    return requests
